use serde_json::Number;
use std::fmt::{Display, Formatter};
use std::sync::Arc;

pub mod device;
pub mod emu;

#[derive(Clone, Debug)]
pub struct BroadcastMessage {
    pub id: String,
    pub key: String,
    pub value: Arc<dyn std::any::Any + Send + Sync>,
}

impl Display for BroadcastMessage {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        if let Some(val) = self.value.downcast_ref::<Number>() {
            write!(f, "({}, {})", self.key, val)
        } else if let Some(val) = self.value.downcast_ref::<String>() {
            write!(f, "({}, {})", self.key, val)
        } else {
            write!(f, "({}, {:?})", self.key, self.value.type_id())
        }
    }
}

/// ModbusTCP开启连接，如果失败重试
/// 初始等待重连时间1秒，每次递增，最大60秒
#[macro_export]
macro_rules! connect_or_retry {
    ($connect_expr:expr,$label:expr) => {{
        use std::cmp::min;
        use tokio::time::{sleep, Duration};
        let mut retry_delay = Duration::from_secs(1);
        let max_delay = Duration::from_secs(60);
        loop {
            match $connect_expr.await {
                Ok(conn) => {
                    log::info!("[{}] 连接成功!", $label);
                    break Ok(conn);
                }
                Err(e) => {
                    log::error!(
                        "[{}] 连接失败: {}, 将在 {:?} 后重试...",
                        $label,
                        e,
                        retry_delay
                    );
                    sleep(retry_delay).await;
                    retry_delay = min(retry_delay * 2, max_delay);
                }
            }
        }
    }};
}
