use crate::internal::modbus::code::ModbusCode;
use serde::{Deserialize, Serialize};

pub mod code;

///最长寄存器字节数量
pub const MAX_WORD_CNT: usize = 120;
///最长寄存器位数量
pub const MAX_BIT_CNT: usize = 2000;

/// CSV对应的结构体
#[derive(Debug, Serialize, Deserialize)]
pub struct CsvData {
    pub address: Option<u16>,
    pub code: Option<u8>,
    pub name: Option<String>,
    #[serde(rename = "type")]
    pub t: Option<String>,
    pub rw: Option<String>,
    pub factor: Option<f32>,
    pub offset: Option<i16>,
}

/// 从路径读取csv文档并解析为集合
/// [path] 文件路径
/// [Vec<ModbusCode>] 原始配置集合
pub fn read_csv_to_code(path: &str) -> anyhow::Result<Vec<ModbusCode>> {
    let mut rdr = csv::Reader::from_path(path)?;
    let mut vec = Vec::new();
    for result in rdr.deserialize() {
        let csv_data: CsvData = result?;
        if let Ok(code) = csv_data.try_into() {
            vec.push(code);
        }
    }
    Ok(vec)
}

/// 描述了基础ModbusCode
pub trait BaseModbusCode {
    fn addr(&self) -> u16;
}

/// 将序列按照连续子序列分组
/// 如: [1,3,2,7,8,11,10] 分割为 [[1,2,3],[7,8],[10,11]]
pub fn slice_sequential<T>(codes: &mut [T], max_count: usize) -> Vec<&mut [T]>
where
    T: BaseModbusCode,
{
    codes.sort_by(|a, b| a.addr().cmp(&b.addr()));
    let mut result = Vec::new();
    let mut remaining = codes;
    while !remaining.is_empty() {
        let mut len = 1;
        while len < remaining.len() {
            let prev_addr = remaining[len - 1].addr();
            let curr_addr = remaining[len].addr();
            if prev_addr + 1 == curr_addr && len < max_count {
                len += 1;
            } else {
                break;
            }
        }
        let (group, rest) = remaining.split_at_mut(len);
        result.push(group);
        remaining = rest;
    }
    result
}

/// 将序列按照连续子序列分组
/// 如: [1,3,2,7,8,11,10] 分割为 [[1,2,3],[7,8],[10,11]]
pub fn sequential<T: BaseModbusCode>(mut codes: Vec<T>, max_count: usize) -> Vec<Vec<T>> {
    //按地址升序排序
    codes.sort_by(|a, b| a.addr().cmp(&b.addr()));
    if codes.is_empty() {
        return Vec::new();
    }
    let mut vec = Vec::new();
    let mut current_group = Vec::new();
    let mut iter = codes.into_iter();
    if let Some(prev) = iter.next() {
        current_group.push(prev);
        for current in iter {
            if let Some(prev) = current_group.last() {
                let prev_addr = prev.addr();
                let curr_addr = current.addr();
                if prev_addr + 1 == curr_addr {
                    if current_group.len() >= max_count {
                        vec.push(current_group);
                        current_group = Vec::new();
                    }
                    current_group.push(current);
                } else {
                    vec.push(current_group);
                    current_group = Vec::new();
                    current_group.push(current)
                }
            }
        }
        vec.push(current_group);
    }
    vec
}

#[cfg(test)]
mod test {
    use super::*;

    struct TestModbusCode(u16);

    impl BaseModbusCode for TestModbusCode {
        fn addr(&self) -> u16 {
            self.0
        }
    }

    #[test]
    fn test_sequential() {
        let base = TestModbusCode { 0: 1u16 };
        let mut codes = vec![
            base,
            TestModbusCode { 0: 3u16 },
            TestModbusCode { 0: 2u16 },
            TestModbusCode { 0: 7u16 },
            TestModbusCode { 0: 8u16 },
            TestModbusCode { 0: 11u16 },
            TestModbusCode { 0: 10u16 },
        ];
        let vec = slice_sequential(codes.as_mut_slice());
        let vec = vec
            .into_iter()
            .map(|it| it.into_iter().map(|i| i.0).collect::<Vec<u16>>())
            .collect::<Vec<Vec<u16>>>();
        assert_eq!(vec, vec![vec![1, 2, 3], vec![7, 8], vec![10, 11],]);
    }
}
